from anneal.core.components import *


class MHChain(BaseChainSA):
    def __init__(self, ObjFunc: ObjectiveFunction, Temperature, Nsim=5000):
        super().__init__(ObjFunc, Temperature, Nsim)
        self.states[0] = np.zeros(ObjFunc.limits.dims)
        self.status[0] = True
        self.cand = None
        self.cur_state = self.states[0]
        self.best = None

    def step(self):
        """Here we pick a candidate which is chain-independent"""
        for step in range(1, self.Nsim):
            cand = self.ObjFunc.limits.mkpoint()
            rho = self.bounds_eval_rho(cand, step - 1)
            if self.proposal() < rho:
                self.states[step] = self.states[step - 1] + (
                    cand - self.states[step]
                )
                self.status[step] = True
            else:
                self.states[step] = self.states[step - 1]
                self.status[step] = False

    def bounds_eval_rho(self, cand, chain_id):
        try:
            val = self.TargetDistrib(self.states[chain_id])
        except OutOfBounds:
            self.states[chain_id] = self.ObjFunc.limits.clip(
                self.states[chain_id]
            )
            val = self.TargetDistrib(self.states[chain_id])
        return self.TargetDistrib(cand) / val

    def proposal(self):
        return np.random.uniform(0, 1, 1)

    def __call__(self):
        self.step()
        # TODO: Why is that last one out of bounds?
        energies = np.array([self.ObjFunc(x) for x in self.states[:-1]])
        min_state = self.states[energies.argmin()]
        self.best = FPair(pos=min_state, val=np.min(energies))
        return self.best



import pytest
import numpy.testing as nptt

from anneal.funcs.obj2d import *
from anneal.quenchers.boltzmann import BoltzmannQuencher


def test_BQST2d():
    ff = StybTang2d()
    bq = BoltzmannQuencher(StybTang2d(), T_init=5, pos_init=np.ones(2) * -2)
    bq()
    assert bq.fCalls < 120000



import pytest
import numpy.testing as nptt

from anneal.funcs.obj2d import *
from anneal.funcs.objNd import *


def test_StybTang2d():
    ff = StybTang2d()
    assert ff(np.array([-5, -5])) == 200
    assert ff(np.array([-4.9, -4.9, -5, -5])) == pytest.approx(
        np.array([167.8201, 200.0])
    )
    assert ff(ff.globmin.pos) == pytest.approx(ff.globmin.val, abs=1e-3)
    assert ff(
        np.array(
            [
                [0.61520743, 0.73489428],
                [0.42910286, 0.02296883],
                [0.33404362, 0.29842335],
            ]
        )
    ) == pytest.approx(
        np.array(
            [-3.755682790653861, -0.3301234957151148, -0.01377451063767443]
        )
    )


def test_StybTangNd_two():
    ffTwo = StybTangNd(dims=2)
    assert ffTwo(np.array([-5, -5])) == 200
    assert ffTwo(np.array([-4.9, -4.9, -5, -5])) == pytest.approx(
        np.array([167.8201, 200.0])
    )
    assert ffTwo(ffTwo.globmin.pos) == pytest.approx(
        ffTwo.globmin.val, abs=1e-3
    )


def test_StybTangNd():
    ff = StybTangNd(dims=5)
    assert ff(
        np.array([0.61520743, 0.73489428, 0.2, 0.6, 0.21])
    ) == pytest.approx(-4.71691038565)
    assert ff(ff.globmin.pos) == pytest.approx(ff.globmin.val, abs=1e-3)
    assert str(ff) == "5D Styblinski-Tang"



from anneal.core.components import NumLimit, ObjectiveFunction, FPair

import numpy as np


class StybTang2d(ObjectiveFunction):
    def __init__(
        self, limits=NumLimit(dims=2, low=np.ones(2) * -5, high=np.ones(2) * 5)
    ):
        super().__init__(
            limits,
            FPair(val=-39.16599 * 2, pos=np.array([-2.903534, -2.903534])),
        )

    def singlepoint(self, pos):
        self.limits.check(pos)
        return np.sum((pos**4) - (16 * (pos**2)) + (5 * pos)) / 2

    def multipoint(self, pos):
        return np.apply_along_axis(
            self.singlepoint, 1, pos.reshape(-1, self.limits.dims)
        )

    def __repr__(self):
        return "2D Styblinski-Tang"



import pytest
import numpy.testing as nptt

from anneal.funcs.obj2d import *
from anneal.mcsamplers.chains import MHChain


def test_MHChain():
    ff = StybTang2d()
    mha = MHChain(StybTang2d(), Temperature=10, Nsim=9000)
    bpair = mha()
    assert bpair.val == pytest.approx(ff.globmin.val, 1e-2)
    assert bpair.pos == pytest.approx(ff.globmin.pos, 1e1)



import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.ticker import LinearLocator, FormatStrFormatter
import numpy as np

from anneal.funcs.obj2d import *
from anneal.core.components import AcceptStates, Quencher


class Plot2dObj:
    """Meant to plot 2D objects"""

    def __init__(self, obj: ObjectiveFunction, nelem: int):
        self.func = obj
        self.nelem = nelem
        fll = self.func.limits.low
        fhh = self.func.limits.high
        step_size_x = abs(fll[0] - fhh[0]) / self.nelem
        step_size_y = abs(fll[1] - fhh[1]) / self.nelem
        plX = np.arange(fll[0], fhh[0], step_size_x)
        plY = np.arange(fll[1], fhh[1], step_size_y)
        self.X, self.Y = np.meshgrid(plX, plY, indexing="xy")
        self.Z = self.prepVals()
        self.contourExtent = [
            np.min(self.X.ravel()),
            np.max(self.X.ravel()),
            np.min(self.Y.ravel()),
            np.max(self.Y.ravel()),
        ]
        self.X_glob_min = self.X.ravel()[self.Z.argmin()]
        self.Y_glob_min = self.Y.ravel()[self.Z.argmin()]
        self.Z_glob_min = np.min(self.Z.ravel())
        # Set to the current grid min
        # TODO: Handle degenerate minima
        if isinstance(obj.globmin, type(None)):
            obj.globmin = FPair(
                val=self.Z_glob_min,
                pos=np.array([self.X_glob_min, self.Y_glob_min]),
            )
        self.pdat = None

    def prepVals(self):
        grid_vals = [
            self.func(np.column_stack([self.X[itera], self.Y[itera]]))
            for itera in range(self.nelem)
        ]
        return np.array(grid_vals)

    def create3d(self, showGlob=True, savePath=None):
        fig = plt.figure(figsize=(12, 10))
        ax = plt.subplot(projection="3d")
        surf = ax.plot_surface(
            self.X, self.Y, self.Z, cmap=cm.coolwarm, alpha=0.7
        )
        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter("%.1f"))
        [t.set_va("center") for t in ax.get_yticklabels()]
        [t.set_ha("left") for t in ax.get_yticklabels()]
        [t.set_va("center") for t in ax.get_xticklabels()]
        [t.set_ha("right") for t in ax.get_xticklabels()]
        [t.set_va("center") for t in ax.get_zticklabels()]
        [t.set_ha("left") for t in ax.get_zticklabels()]
        fig.colorbar(surf, shrink=0.35, aspect=8)
        ax.view_init(elev=15, azim=220)
        if showGlob:
            ax.scatter(
                self.X_glob_min,
                self.Y_glob_min,
                self.Z_glob_min,
                color="black",
                alpha=1,
            )
            ax.text(
                self.X_glob_min,
                self.Y_glob_min,
                self.Z_glob_min,
                "Global Minima",
                color="black",
                alpha=1,
            )
        plt.xlabel("X")
        plt.ylabel("Y")
        plt.title(str(self.func))
        if savePath is not None:
            plt.savefig(savePath, dpi=300)
        else:
            plt.show()

    def createContour(self, showGlob=True, savePath=None):
        fig = plt.figure(figsize=(12, 10))
        ax = plt.subplot()
        [t.set_va("center") for t in ax.get_yticklabels()]
        [t.set_ha("left") for t in ax.get_yticklabels()]
        [t.set_va("center") for t in ax.get_xticklabels()]
        [t.set_ha("right") for t in ax.get_xticklabels()]
        plt.imshow(
            self.Z,
            extent=self.contourExtent,
            origin="lower",
            cmap="viridis",
            alpha=0.8,
        )
        plt.colorbar()
        contours = ax.contour(
            self.X, self.Y, self.Z, 10, colors="black", alpha=0.9
        )
        plt.clabel(contours, inline=True, fontsize=8, fmt="%.0f")
        if showGlob:
            plt.plot(
                self.X_glob_min,
                self.Y_glob_min,
                color="white",
                marker="x",
                markersize=5,
            )
            ax.text(
                self.X_glob_min + 0.1,
                self.Y_glob_min,
                "Global Minima",
                color="white",
            )
        if savePath is not None:
            plt.savefig(savePath, dpi=300)
        else:
            plt.show()

    def plotQuenchContour(self, quenchy: Quencher, nsamples=500, savePath=None):
        self.pdat = pd.DataFrame(quenchy.PlotData)
        accepted_pos = self.pdat[self.pdat.accept == AcceptStates.IMPROVED][
            "pos"
        ]
        rejected_pos = self.pdat[self.pdat.accept == AcceptStates.REJECT]["pos"]
        mhaccept_pos = self.pdat[self.pdat.accept == AcceptStates.MHACCEPT][
            "pos"
        ]
        toNPD = (
            lambda poscol: np.concatenate(poscol.to_list(), axis=0).reshape(
                -1, self.func.limits.dims
            )
            if len(poscol) > 1
            else poscol
        )
        getDat = (
            lambda posDat: toNPD(pd.DataFrame.sample(posDat, n=nsamples))
            if len(posDat) > nsamples
            else toNPD(posDat)
        )
        inBounds = lambda point: np.all(
            point > self.contourExtent[::2]
        ) and np.all(point < self.contourExtent[1::2])
        fig = plt.figure(figsize=(12, 10))
        ax = plt.subplot()
        plotAccept = (
            lambda point: ax.plot(
                point[0], point[1], marker="o", color="blue", alpha=0.5
            )
            if inBounds(point)
            else None
        )
        plotReject = (
            lambda point: ax.plot(
                point[0], point[1], marker="o", color="red", alpha=0.3
            )
            if inBounds(point)
            else None
        )
        plotMHAccept = (
            lambda point: ax.plot(
                point[0],
                point[1],
                marker="*",
                color="yellow",
                alpha=0.5,
            )
            if inBounds(point)
            else None
        )
        [t.set_va("center") for t in ax.get_yticklabels()]
        [t.set_ha("left") for t in ax.get_yticklabels()]
        [t.set_va("center") for t in ax.get_xticklabels()]
        [t.set_ha("right") for t in ax.get_xticklabels()]
        plt.imshow(
            self.Z,
            extent=self.contourExtent,
            origin="lower",
            cmap="viridis",
            alpha=0.8,
        )
        plt.colorbar()
        contours = ax.contour(
            self.X, self.Y, self.Z, 10, colors="black", alpha=0.9
        )
        plt.clabel(contours, inline=True, fontsize=8, fmt="%.0f")
        plt.plot(
            self.X_glob_min,
            self.Y_glob_min,
            color="white",
            marker="x",
            markersize=5,
        )
        if inBounds(quenchy.best.pos):
            ax.plot(
                quenchy.best.pos[0],
                quenchy.best.pos[1],
                marker="*",
                color="white",
                alpha=1,
            )
        else:
            ax.plot(
                quenchy.best.pos[0],
                quenchy.best.pos[1],
                marker="*",
                color="black",
                alpha=1,
            )
        for point in getDat(accepted_pos):
            plotAccept(point)
        for point in getDat(rejected_pos):
            plotReject(point)
        for point in getDat(mhaccept_pos):
            plotMHAccept(point)
        if savePath is not None:
            plt.savefig(savePath, dpi=300)
        else:
            plt.show()



from dataclasses import dataclass
from collections import namedtuple
from enum import Enum
import abc

import numpy as np
import numpy.typing as npt
import typing

from anneal.core.exceptions import OutOfBounds

MAX_LIMITS = namedtuple("MAX_LIMITS", ["EPOCHS", "STEPS_PER_EPOCH"])
EpochLine = namedtuple(
    "EpochLine", ["epoch", "temperature", "step", "pos", "val", "accept"]
)
AcceptStates = Enum("AcceptStates", ["IMPROVED", "MHACCEPT", "REJECT"])


@dataclass
class FPair:
    """Class for handing position value pairs"""

    pos: npt.NDArray
    val: float

    def EvalFunc(self, ObjFunc: callable):
        self.val = ObjFunc(self.pos)


@dataclass
class NumLimit:
    """Class for tracking function bounds

    Parameters
    ----------
    low: A list of values which the function must not be below
    high: A list of values which the function must not exceed
    slack: The amount by which bounds may be exceeded without an error
    dims: This should be the same as the length of the list
    """

    low: npt.NDArray
    high: npt.NDArray
    slack: float = 1e-6
    dims: int = 1

    def check(self, pos: npt.NDArray):
        if not (
            np.all(pos > self.low - self.slack)
            and np.all(pos < self.high + self.slack)
        ):
            raise OutOfBounds(
                f"{pos} is not within {self.slack} of {self.low} and {self.high}"
            )
        return

    def mkpoint(self) -> npt.NDArray:
        """Generate a random point

        TODO: Handle other constraints (undefined regions)
        """
        return np.random.default_rng().uniform(self.low, self.high, self.dims)

    def clip(self, point: npt.NDArray) -> npt.NDArray:
        """Clips values"""
        return np.clip(point, self.low, self.high)


class ObjectiveFunction(metaclass=abc.ABCMeta):
    def __init__(self, limits: NumLimit, global_min: FPair = None):
        self.calls = 0
        self.limits = limits
        self.globmin = global_min

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "__call__")
            and callable(subclass.__call__)
            and hasattr(subclass, "pointwise")
            and callable(subclass.pointwise)
            and hasattr(subclass, "multipoint")
            and callable(subclass.multipoint)
            and hasattr(subclass, "__repr__")
            and callable(subclass.__repr__)
            or NotImplemented
        )

    def __call__(self, pos):
        ## TODO: calls in multipoint may be more than once
        if pos.ravel().shape[0] != self.limits.dims:
            self.calls += 1
            return self.multipoint(pos)
        else:
            self.calls += 1
            return self.singlepoint(pos)

    @abc.abstractmethod
    def singlepoint(self, pos):
        """Evaluate the function at a single configuration"""
        raise NotImplementedError(
            "Need to be able to call the objective function on a single point"
        )

    @abc.abstractmethod
    def multipoint(self, pos):
        """Evaluate the function at many configurations

        TODO: This allows for a faster implementation in C
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        """Name the function"""
        raise NotImplementedError


class AcceptCriteria(metaclass=abc.ABCMeta):
    """The acceptance criteria for selecting an unfavorable move"""

    def __init__(self):
        pass

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "__call__")
            and callable(subclass.__call__)
            and hasattr(subclass, "__repr__")
            and callable(subclass.__repr__)
            or NotImplemented
        )

    @abc.abstractmethod
    def __call__(self):
        """Accept or reject"""
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        """Name the function"""
        raise NotImplementedError


class ConstructNeighborhood(metaclass=abc.ABCMeta):
    """Choosing a feasible point / direction

    For most methods, this is a uniform distribution
    """

    def __init__(self):
        pass

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "__call__")
            and callable(subclass.__call__)
            and hasattr(subclass, "__repr__")
            and callable(subclass.__repr__)
            or NotImplemented
        )

    @abc.abstractmethod
    def __call__(self):
        """Yield a point"""
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        """Name the function"""
        raise NotImplementedError


class CoolingSchedule(metaclass=abc.ABCMeta):
    def __init__(self):
        pass

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "__call__")
            and callable(subclass.__call__)
            and hasattr(subclass, "__repr__")
            and callable(subclass.__repr__)
            or NotImplemented
        )

    @abc.abstractmethod
    def __call__(self):
        """Generate a temperature"""
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        """Name the function"""
        raise NotImplementedError


class MoveClass(metaclass=abc.ABCMeta):
    """The probability distribution for the step-size

    Also called the visiting distribution.
    """

    def __init__(self):
        pass

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "__call__")
            and callable(subclass.__call__)
            and hasattr(subclass, "__repr__")
            and callable(subclass.__repr__)
            or NotImplemented
        )

    @abc.abstractmethod
    def __call__(self):
        """Generate a move"""
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        """Name the function"""
        raise NotImplementedError


class Quencher(metaclass=abc.ABCMeta):
    """A quenching class"""

    def __init__(
        self,
        ObjFunc: ObjectiveFunction,
        MkNeigh: ConstructNeighborhood,
        Accepter: AcceptCriteria,
        Mover: MoveClass,
        Cooler: CoolingSchedule,
    ):
        self.ObjFunc = ObjFunc
        self.MkNeigh = MkNeigh
        self.Accepter = Accepter
        self.Mover = Mover
        self.Cooler = Cooler
        self.cur = FPair(None, None)
        self.candidate = FPair(None, None)
        self.best = FPair(None, None)
        self.acceptances = 0
        self.rejections = 0
        self.samestate_time = 0
        self.fCalls = 0
        self.PlotData = []

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "__call__")
            and callable(subclass.__call__)
            and hasattr(subclass, "__repr__")
            and callable(subclass.__repr__)
            and hasattr(subclass, "HasConverged")
            and callable(subclass.HasConverged)
            or NotImplemented
        )

    @abc.abstractmethod
    def __call__(self):
        """Generate a move"""
        raise NotImplementedError

    @abc.abstractmethod
    def HasConverged(self):
        """Check for Convergence"""
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        """Name the function"""
        raise NotImplementedError


class BaseChainSA(metaclass=abc.ABCMeta):
    """The Base chain for MCMC samplers used in SA"""

    def __init__(
        self,
        ObjFunc: ObjectiveFunction,
        Temperature,
        Nsim=5000,
    ):
        """A key point is that the temperature is used to define the target probability distribution"""
        self.ObjFunc = ObjFunc
        self.Nsim = Nsim
        self.Temperature = Temperature
        self.stepNum = 0
        self.states = np.array(
            [self.ObjFunc.limits.mkpoint() for x in range(self.Nsim)]
        )
        self.status = np.empty(self.Nsim, dtype=bool)
        self.fvals = np.zeros(self.Nsim)
        ## This is the probability at each temperature
        self.TargetDistrib = lambda point: np.exp(
            -self.ObjFunc(point) / self.Temperature
        )



from anneal.core.components import NumLimit, ObjectiveFunction, FPair

import numpy as np


class StybTangNd(ObjectiveFunction):
    """This is a generalization as the function is evaluated on a hypercube"""

    def __init__(self, dims):
        self.dims = dims
        self.limits = NumLimit(
            dims=self.dims,
            low=np.ones(self.dims) * -5,
            high=np.ones(self.dims) * 5,
        )
        super().__init__(
            self.limits,
            FPair(
                val=-39.16599 * self.dims, pos=np.array([-2.903534] * self.dims)
            ),
        )

    def singlepoint(self, pos):
        self.limits.check(pos)
        return np.sum((pos**4) - (16 * (pos**2)) + (5 * pos)) / 2

    def multipoint(self, pos):
        return np.apply_along_axis(
            self.singlepoint, 1, pos.reshape(-1, self.limits.dims)
        )

    def __repr__(self):
        return f"{self.dims}D Styblinski-Tang"



import numpy as np
import numpy.typing as npt
import pytest

from anneal.core.components import (
    Quencher,
    ObjectiveFunction,
    CoolingSchedule,
    AcceptCriteria,
    ConstructNeighborhood,
    MoveClass,
    NumLimit,
    MAX_LIMITS,
    EpochLine,
    FPair,
    AcceptStates,
)


class BoltzmannQuencher(Quencher):
    def __init__(
        self,
        ObjFunc: ObjectiveFunction,
        pos_init: npt.NDArray = None,
        *,
        T_init: float = None,
        maxiter: MAX_LIMITS = MAX_LIMITS(int(1e6), int(1e3)),
    ):
        self.ObjFunc = ObjFunc
        self.Mover = BoltzmannMove(self.ObjFunc.limits)
        self.MkNeigh = BoltzmannNeighbor(self.ObjFunc.limits, self.Mover)
        self.Accepter = BoltzmannAccept()
        ## Maybe Tinit = −∆E / ln x0
        self.T_init = T_init
        self.Cooler = BoltzmannCooler(self.T_init)
        super().__init__(
            self.ObjFunc,
            self.MkNeigh,
            self.Accepter,
            self.Mover,
            self.Cooler,
        )
        self.epoch = 1
        if pos_init is None:
            self.cur = FPair(
                pt := self.ObjFunc.limits.mkpoint(), self.ObjFunc(pt)
            )
        else:
            self.cur = FPair(pt := pos_init, self.ObjFunc(pt))
        self.maxiter = maxiter
        ## Initially
        self.best = self.cur

    def __call__(self, trackPlot=False):
        while (
            temperature := self.Cooler(self.epoch)
        ) > 0.1 and self.epoch < self.maxiter.EPOCHS:
            for step in range(1, self.maxiter.STEPS_PER_EPOCH + 1):
                self.candidate = FPair(
                    pt := self.MkNeigh(self.cur.pos), self.ObjFunc(pt)
                )
                diff = self.DeltaFunc()
                if diff <= 0:
                    ## Better point, accept always
                    self.AcceptMove()
                    self.PlotData.append(
                        EpochLine(
                            epoch=self.epoch,
                            temperature=temperature,
                            step=step,
                            pos=self.candidate.pos,
                            val=self.candidate.val,
                            accept=AcceptStates.IMPROVED,
                        )
                    )
                else:
                    if self.Accepter(diff, temperature):
                        self.AcceptMove()
                        self.PlotData.append(
                            EpochLine(
                                epoch=self.epoch,
                                temperature=temperature,
                                step=step,
                                pos=self.candidate.pos,
                                val=self.candidate.val,
                                accept=AcceptStates.MHACCEPT,
                            )
                        )
                    else:
                        self.RejectMove()
                        self.PlotData.append(
                            EpochLine(
                                epoch=self.epoch,
                                temperature=temperature,
                                step=step,
                                pos=self.candidate.pos,
                                val=self.candidate.val,
                                accept=AcceptStates.REJECT,
                            )
                        )
            print(f"{self.epoch} for {temperature} has {self.best}")
            if self.HasConverged():
                return
            else:
                self.epoch += 1

    def DeltaFunc(self):
        return self.candidate.val - self.cur.val

    def AcceptMove(self):
        self.cur = self.candidate
        if self.cur.val < self.best.val:
            self.best = self.cur
        else:
            self.samestate_time += 1
        self.acceptances += 1

    def RejectMove(self):
        self.rejections += 1
        self.samestate_time += 1

    # TODO: Generalize for other kinds of convergence
    def HasConverged(self):
        if self.best.val == pytest.approx(
            self.ObjFunc.globmin.val, 1e-3
        ) or self.best.pos == pytest.approx(self.ObjFunc.globmin.pos, 1e-3):
            self.fCalls = self.ObjFunc.calls
            self.ObjFunc.calls = 0
            return True

    def __repr__(self):
        return "An instance of the Boltzmann Quencher"


class BoltzmannCooler(CoolingSchedule):
    """This is the log schedule of Geman and Geman (1983)"""

    def __init__(self, T_init, c_param=1):
        """Salazar and Torale (1997) set c_param to T_init"""
        self.Tinit = T_init
        self.c_param = c_param

    def __call__(self, epoch):
        """Random perturbation to ensure G&G inequality holds
        np.random.uniform(0.01, 0.5) +
        """
        Temp = (self.c_param * self.Tinit) / (1 + np.log(epoch))
        return Temp

    def __repr__(self):
        return "Log schedule cooler for Boltzmann"


class BoltzmannMove(MoveClass):
    """For the Boltzmann the stepsize is random
    Here we take a standard Gaussian distribution
    """

    def __init__(self, limits: NumLimit):
        self.limits = limits

    def __call__(self):
        return np.random.default_rng().normal()

    def __repr__(self):
        return "Random stepsize for the Boltzmann"


## Consider np.random.normal(0, np.random.rand(1), ndim)
class BoltzmannNeighbor(ConstructNeighborhood):
    """For the Boltzmann this is a random direction, Casella and Robert (2010)
    Gall (2014) disagrees though
    """

    def __init__(self, limits: NumLimit, mover: BoltzmannMove):
        self.limits = limits
        self.stepper = mover

    def __call__(self, c_pos):
        ## Project current onto candidate
        candidate = self.limits.mkpoint()
        proj = (
            candidate * np.dot(c_pos, candidate) / np.dot(candidate, candidate)
        )
        dir_vec = proj / np.linalg.norm(proj)
        final_cand = candidate + dir_vec * self.stepper()
        ## Clipping is not part of the original formulation
        # final_cand = self.limits.clip(final_cand)
        return final_cand

    def __repr__(self):
        return "Random candidate for the Boltzmann"


class BoltzmannAccept(AcceptCriteria):
    def __init__(self, k=1):
        self.k = 1

    def __call__(self, diff, Temperature):
        metropolis = np.min([np.exp(-self.k * diff / Temperature), 1])
        return np.random.uniform(0, 1) < metropolis

    def __repr__(self):
        return "Basic Metropolis acceptance for Boltzmann"
